/* gtkspringlayout.c
 * 
 * Copyright (c) 2005 Red Hat, Inc.
 * Author Matthias Clasen
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gtkspringlayout.h"

#define P_(x) (x)

typedef struct _GtkSpringLayoutChild GtkSpringLayoutChild;
struct _GtkSpringLayoutChild
{
  GtkWidget *widget;
  
  GtkSpring *x;
  GtkSpring *y;
  GtkSpring *width;
  GtkSpring *height;
  GtkSpring *right;
  GtkSpring *bottom;

  GtkSpring *h;
  GtkSpring *v;

  gint extra_width;
  gint max_extra_width;
  gint extra_height;
  gint max_extra_height;
};

static GtkSpring *
gtk_spring_layout_child_get_constraint (GtkSpringLayoutChild *child,
					GtkConstraintType     pos);


/* GtkSpring */

enum 
{
  GTK_SPRING_CYCLE_UNSET,
  GTK_SPRING_CYCLIC,
  GTK_SPRING_ACYCLIC
};

G_DEFINE_ABSTRACT_TYPE(GtkSpring, gtk_spring, G_TYPE_OBJECT)

static gboolean 
gtk_spring_real_is_cyclic (GtkSpring *spring)
{
  return FALSE;
}

static void
gtk_spring_class_init (GtkSpringClass *class)
{
  class->is_cyclic = gtk_spring_real_is_cyclic;
}

static void
gtk_spring_init (GtkSpring *spring)
{
  spring->min = GTK_SPRING_UNSET;
  spring->max = GTK_SPRING_UNSET;
  spring->preferred = GTK_SPRING_UNSET;
  spring->value = GTK_SPRING_UNSET;

  spring->cyclic = GTK_SPRING_CYCLE_UNSET;
  spring->floating = TRUE;
}

static GType gtk_simple_spring_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_SIMPLE_SPRING (gtk_simple_spring_get_type ())
#define GTK_SIMPLE_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_CAST ((object), GTK_TYPE_SIMPLE_SPRING, GtkSimpleSpring))
#define GTK_IS_SIMPLE_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_TYPE ((object), GTK_TYPE_SIMPLE_SPRING))

typedef GtkSpring GtkSimpleSpring;
typedef GtkSpringClass GtkSimpleSpringClass;

G_DEFINE_TYPE(GtkSimpleSpring, gtk_simple_spring, GTK_TYPE_SPRING)

static gint 
gtk_simple_spring_get_minimum (GtkSpring *spring)
{
  return spring->min;
}

static gint 
gtk_simple_spring_get_preferred (GtkSpring *spring)
{
  return spring->preferred;  
}

static gint 
gtk_simple_spring_get_maximum (GtkSpring *spring)
{
  return spring->max;  
}

static gint 
gtk_simple_spring_get_value (GtkSpring *spring)
{
  if (spring->value == GTK_SPRING_UNSET)
    spring->value = spring->preferred;

  return spring->value;
}

static void
gtk_simple_spring_set_value (GtkSpring *spring,
			     gint       value)
{
  spring->value = value;
}

static void
gtk_simple_spring_class_init (GtkSimpleSpringClass *class)
{
  GtkSpringClass *spring_class = GTK_SPRING_CLASS (class);

  spring_class->get_minimum_value = gtk_simple_spring_get_minimum;
  spring_class->get_preferred_value = gtk_simple_spring_get_preferred;
  spring_class->get_maximum_value = gtk_simple_spring_get_maximum;
  spring_class->get_value = gtk_simple_spring_get_value;
  spring_class->set_value = gtk_simple_spring_set_value;
}

static void
gtk_simple_spring_init (GtkSpring *spring)
{
}

static GType gtk_combine_spring_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_COMBINE_SPRING (gtk_combine_spring_get_type ())
#define GTK_COMBINE_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_CAST ((object), GTK_TYPE_COMBINE_SPRING, GtkCombineSpring))
#define GTK_IS_COMBINE_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_TYPE ((object), GTK_TYPE_COMBINE_SPRING))

typedef struct _GtkCombineSpring GtkCombineSpring;
typedef struct _GtkCombineSpringClass GtkCombineSpringClass;

enum {
  GTK_COMBINE_SPRING_ADD,
  GTK_COMBINE_SPRING_MAX,
  GTK_COMBINE_SPRING_DIFFERENCE
};

struct _GtkCombineSpring
{
  GtkSpring spring;
  
  GtkSpring *spring1;
  GtkSpring *spring2;

  gint op;
};

struct _GtkCombineSpringClass
{
  GtkSpringClass parent_class;
};

G_DEFINE_TYPE(GtkCombineSpring, gtk_combine_spring, GTK_TYPE_SPRING)

static gint
combine (GtkCombineSpring *spring,
	 gint              v1,
	 gint              v2)
{
  gint res;

  switch (spring->op)
    {
    case GTK_COMBINE_SPRING_ADD:
      res = v1 + v2;
      break;
    case GTK_COMBINE_SPRING_MAX:
      res = MAX (v1, v2);
      break;
    case GTK_COMBINE_SPRING_DIFFERENCE:
      res = v1 - v2;
      break;
    }

  return res;
}

static gint 
gtk_combine_spring_get_minimum (GtkSpring *spring)
{
  if (spring->min == GTK_SPRING_UNSET)
    {
      GtkCombineSpring *cs = GTK_COMBINE_SPRING (spring);
      gint v1, v2;
      
      v1 = gtk_spring_get_minimum (cs->spring1);
      v2 = gtk_spring_get_minimum (cs->spring2);
      spring->min = combine (cs, v1, v2);
    }

  return spring->min;
}

static gint 
gtk_combine_spring_get_preferred (GtkSpring *spring)
{
  if (spring->preferred == GTK_SPRING_UNSET)
    {
      GtkCombineSpring *cs = GTK_COMBINE_SPRING (spring);
      gint v1, v2;
      
      v1 = gtk_spring_get_preferred (cs->spring1);
      v2 = gtk_spring_get_preferred (cs->spring2);
      spring->preferred = combine (cs, v1, v2);
    }

  return spring->preferred;
}

static gint 
gtk_combine_spring_get_maximum (GtkSpring *spring)
{
  if (spring->max == GTK_SPRING_UNSET)
    {
      GtkCombineSpring *cs = GTK_COMBINE_SPRING (spring);
      gint v1, v2;
      
      v1 = gtk_spring_get_maximum (cs->spring1);
      v2 = gtk_spring_get_maximum (cs->spring2);
      spring->max = combine (cs, v1, v2);
    }

  return spring->max;
}

static gint 
gtk_combine_spring_get_value (GtkSpring *spring)
{
  if (spring->value == GTK_SPRING_UNSET)
    {
      GtkCombineSpring *cs = GTK_COMBINE_SPRING (spring);
      gint v1, v2;
      
      v1 = gtk_spring_get_value (cs->spring1);
      v2 = gtk_spring_get_value (cs->spring2);
      spring->value = combine (cs, v1, v2);
    }

  return spring->value;
}

static gint
gtk_spring_get_range (GtkSpring *spring, 
		      gboolean   compressed)
{
  if (compressed)
    return gtk_spring_get_preferred (spring) - gtk_spring_get_minimum (spring);
  else
    return gtk_spring_get_maximum (spring) - gtk_spring_get_preferred (spring);
}

static gdouble 
gtk_spring_get_strain (GtkSpring *spring)
{
  gint v, p, r;

  v = gtk_spring_get_value (spring);
  p = gtk_spring_get_preferred (spring);
  r = gtk_spring_get_range (spring, v < p);

  if (r == 0)
    r = 1;

  return (gdouble)(v - p) / r;
}

static void 
gtk_spring_set_strain (GtkSpring *spring,
		       gdouble    strain)
{
  gtk_spring_set_value (spring, 
			gtk_spring_get_preferred (spring) + strain * gtk_spring_get_range (spring, strain < 0));
}

static void
gtk_combine_spring_set_value (GtkSpring *spring,
			      gint       value)
{
  GtkCombineSpring *cs = GTK_COMBINE_SPRING (spring);

  if (value == GTK_SPRING_UNSET)
    {
      if (spring->value != GTK_SPRING_UNSET)
	{
	  spring->min = GTK_SPRING_UNSET;
	  spring->preferred = GTK_SPRING_UNSET;
	  spring->max = GTK_SPRING_UNSET;
	  
	  spring->value = GTK_SPRING_UNSET;
	  gtk_spring_set_value (cs->spring1, GTK_SPRING_UNSET);
	  gtk_spring_set_value (cs->spring2, GTK_SPRING_UNSET);
	}

      return;
    }

  spring->value = value;
  
  switch (cs->op)
    {
    case GTK_COMBINE_SPRING_ADD:
      {
	gdouble strain;
	
	strain = gtk_spring_get_strain (spring);
	gtk_spring_set_strain (cs->spring1, strain);
	gtk_spring_set_value (cs->spring2, value - gtk_spring_get_value (cs->spring1));      
      }
      break;

    case GTK_COMBINE_SPRING_DIFFERENCE:
      {
	gdouble strain;
	
	strain = gtk_spring_get_strain (spring);
	gtk_spring_set_strain (cs->spring1, strain);
	gtk_spring_set_value (cs->spring2, gtk_spring_get_value (cs->spring1) - value);      
      }
      break;

    case GTK_COMBINE_SPRING_MAX:
      {
	gint p1, p2;
      
	p1 = gtk_spring_get_preferred (cs->spring1);
	p2 = gtk_spring_get_preferred (cs->spring2);
	
	if (p1 < p2)
	  {
	    gtk_spring_set_value (cs->spring1, MIN (value, p1));
	    gtk_spring_set_value (cs->spring2, value);
	  }
	else 
	  {
	    gtk_spring_set_value (cs->spring1, value);
	    gtk_spring_set_value (cs->spring2, MIN (value, p2));
	  }
      }
      break;
    }
}

static gboolean
gtk_spring_test_cycle (GtkSpring *spring)
{
  if (spring->cyclic == GTK_SPRING_CYCLIC)
    return TRUE;
  
  if (spring->cyclic == GTK_SPRING_ACYCLIC)
    return FALSE;

  spring->cyclic = GTK_SPRING_CYCLIC;
  
  if (!gtk_spring_is_cyclic (spring))
    spring->cyclic = GTK_SPRING_ACYCLIC;

  return spring->cyclic == GTK_SPRING_CYCLIC;    
}


static gboolean 
gtk_combine_spring_is_cyclic (GtkSpring *spring)
{
  GtkCombineSpring *cs = GTK_COMBINE_SPRING (spring);

  return gtk_spring_test_cycle (cs->spring1) || 
    gtk_spring_test_cycle (cs->spring2);
}

static void 
gtk_combine_spring_reset_cyclic (GtkSpring *spring)
{
  GtkCombineSpring *cs = GTK_COMBINE_SPRING (spring);

  gtk_spring_reset_cyclic (cs->spring1);
  gtk_spring_reset_cyclic (cs->spring2);
}

static void
gtk_combine_spring_finalize (GObject *object)
{
  GtkCombineSpring *cs = GTK_COMBINE_SPRING (object);

  g_object_unref (cs->spring1);
  g_object_unref (cs->spring2);
}

static void
gtk_combine_spring_class_init (GtkCombineSpringClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);
  GtkSpringClass *spring_class = GTK_SPRING_CLASS (class);

  object_class->finalize = gtk_combine_spring_finalize;

  spring_class->get_minimum_value = gtk_combine_spring_get_minimum;
  spring_class->get_preferred_value = gtk_combine_spring_get_preferred;
  spring_class->get_maximum_value = gtk_combine_spring_get_maximum;
  spring_class->get_value = gtk_combine_spring_get_value;
  spring_class->set_value = gtk_combine_spring_set_value;
  spring_class->is_cyclic = gtk_combine_spring_is_cyclic;
  spring_class->reset_cyclic = gtk_combine_spring_reset_cyclic;
}

static void
gtk_combine_spring_init (GtkCombineSpring *spring)
{
}

static GType gtk_scale_spring_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_SCALE_SPRING (gtk_scale_spring_get_type ())
#define GTK_SCALE_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_CAST ((object), GTK_TYPE_SCALE_SPRING, GtkScaleSpring))
#define GTK_IS_SCALE_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_TYPE ((object), GTK_TYPE_SCALE_SPRING))

typedef struct _GtkScaleSpring GtkScaleSpring;
typedef GtkSpringClass GtkScaleSpringClass;

struct _GtkScaleSpring
{
  GtkSpring spring;
  
  GtkSpring *spring1;
  gdouble factor;
};

G_DEFINE_TYPE(GtkScaleSpring, gtk_scale_spring, GTK_TYPE_SPRING)

static gint 
gtk_scale_spring_get_minimum (GtkSpring *spring)
{
  if (spring->min == GTK_SPRING_UNSET)
    {
      GtkScaleSpring *s = GTK_SCALE_SPRING (spring);
      
      if (s->factor < 0)
	spring->min = s->factor * gtk_spring_get_maximum (s->spring1);
      else
	spring->min = s->factor * gtk_spring_get_minimum (s->spring1);
    }

  return spring->min;
}

static gint 
gtk_scale_spring_get_preferred (GtkSpring *spring)
{
  if (spring->preferred == GTK_SPRING_UNSET)
    {
      GtkScaleSpring *s = GTK_SCALE_SPRING (spring);

      spring->preferred = s->factor * gtk_spring_get_preferred (s->spring1);
    }

  return spring->preferred;
}

static gint 
gtk_scale_spring_get_maximum (GtkSpring *spring)
{
  if (spring->max == GTK_SPRING_UNSET)
    {
      GtkScaleSpring *s = GTK_SCALE_SPRING (spring);
      
      if (s->factor < 0)
	spring->max = s->factor * gtk_spring_get_minimum (s->spring1);
      else
	spring->max = s->factor * gtk_spring_get_maximum (s->spring1);
    }

  return spring->max;
}

static gint 
gtk_scale_spring_get_value (GtkSpring *spring)
{
  if (spring->value == GTK_SPRING_UNSET)
    {
      GtkScaleSpring *s = GTK_SCALE_SPRING (spring);
      
      spring->value = s->factor * gtk_spring_get_value (s->spring1);
    }

  return spring->value;
}

static void
gtk_scale_spring_set_value (GtkSpring *spring,
			    gint       value)
{
  GtkScaleSpring *s = GTK_SCALE_SPRING (spring);

  if (value == GTK_SPRING_UNSET)
    {
      if (spring->value != GTK_SPRING_UNSET)
	{
	  spring->min = GTK_SPRING_UNSET;
	  spring->preferred = GTK_SPRING_UNSET;
	  spring->max = GTK_SPRING_UNSET;
	  
	  spring->value = GTK_SPRING_UNSET;
	  gtk_spring_set_value (s->spring1, GTK_SPRING_UNSET);
	}

      return;
    }

  spring->value = value;
  gtk_spring_set_value (s->spring1, value / s->factor);
}

static void
gtk_scale_spring_finalize (GObject  *object)
{
  GtkScaleSpring *s = GTK_SCALE_SPRING (object);

  g_object_unref (s->spring1);
}

static gboolean 
gtk_scale_spring_is_cyclic (GtkSpring *spring)
{
  GtkScaleSpring *s = GTK_SCALE_SPRING (spring);

  return gtk_spring_is_cyclic (s->spring1);
}

static void 
gtk_scale_spring_reset_cyclic (GtkSpring *spring)
{
  GtkScaleSpring *s = GTK_SCALE_SPRING (spring);

  gtk_spring_reset_cyclic (s->spring1);
}

static void
gtk_scale_spring_class_init (GtkScaleSpringClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);
  GtkSpringClass *spring_class = GTK_SPRING_CLASS (class);

  object_class->finalize = gtk_scale_spring_finalize;

  spring_class->get_minimum_value = gtk_scale_spring_get_minimum;
  spring_class->get_preferred_value = gtk_scale_spring_get_preferred;
  spring_class->get_maximum_value = gtk_scale_spring_get_maximum;
  spring_class->get_value = gtk_scale_spring_get_value;
  spring_class->set_value = gtk_scale_spring_set_value;
  spring_class->is_cyclic = gtk_scale_spring_is_cyclic;
  spring_class->reset_cyclic = gtk_scale_spring_reset_cyclic;
}

static void
gtk_scale_spring_init (GtkScaleSpring *spring)
{
}

static GType gtk_widget_spring_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_WIDGET_SPRING (gtk_widget_spring_get_type ())
#define GTK_WIDGET_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_CAST ((object), GTK_TYPE_WIDGET_SPRING, GtkWidgetSpring))
#define GTK_IS_WIDGET_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_TYPE ((object), GTK_TYPE_WIDGET_SPRING))

typedef struct _GtkWidgetSpring GtkWidgetSpring;
typedef GtkSpringClass  GtkWidgetSpringClass;

struct _GtkWidgetSpring
{
  GtkSpring spring;
  
  GtkWidget *widget;
  GtkSpringLayoutChild *child;

  gboolean width;
};

G_DEFINE_TYPE(GtkWidgetSpring, gtk_widget_spring, GTK_TYPE_SPRING)

static GtkSpringLayoutChild *get_child (GtkSpringLayout  *layout,
					GtkWidget        *widget);

static gint 
gtk_widget_spring_get_minimum (GtkSpring *spring)
{
  if (spring->min == GTK_SPRING_UNSET)
    {
      GtkWidgetSpring *ws = GTK_WIDGET_SPRING (spring);

      if (GTK_IS_SPRING_LAYOUT (ws->widget))
	{
	  spring->min = ws->width ? ws->widget->allocation.width 
	    : ws->widget->allocation.height;
	}
      else
	{
	  GtkRequisition requisition;

	  gtk_widget_get_child_requisition (ws->widget, &requisition);
	  
	  spring->min = ws->width ? requisition.width : requisition.height;
	}
    }

  return spring->min;
}

static gboolean 
gtk_widget_spring_ensure_child (GtkWidgetSpring *ws)
{
  if (!GTK_IS_SPRING_LAYOUT (ws->widget->parent))
    ws->child = NULL;
  else if (!ws->child)
    ws->child = get_child (GTK_SPRING_LAYOUT (ws->widget->parent), 
			   ws->widget);

  return ws->child != NULL;
}

static gint 
gtk_widget_spring_get_preferred (GtkSpring *spring)
{
  if (spring->preferred == GTK_SPRING_UNSET)
    {
      GtkWidgetSpring *ws = GTK_WIDGET_SPRING (spring);

      spring->preferred = gtk_widget_spring_get_minimum (spring);

      if (gtk_widget_spring_ensure_child (ws))
	spring->preferred += ws->width ? ws->child->extra_width : ws->child->extra_height;
    }

  return spring->preferred;
}

static gint 
gtk_widget_spring_get_maximum (GtkSpring *spring)
{
  if (spring->max == GTK_SPRING_UNSET)
    {
      GtkWidgetSpring *ws = GTK_WIDGET_SPRING (spring);

      spring->max = gtk_widget_spring_get_minimum (spring);
      
      if (gtk_widget_spring_ensure_child (ws)) 
	spring->max += ws->width ? ws->child->max_extra_width : ws->child->max_extra_height;
      
    }

  return spring->max;
}

static gint 
gtk_widget_spring_get_value (GtkSpring *spring)
{
  if (spring->value == GTK_SPRING_UNSET)
    {
      GtkWidgetSpring *ws = GTK_WIDGET_SPRING (spring);

      spring->value = gtk_widget_spring_get_minimum (spring);
      if (gtk_widget_spring_ensure_child (ws)) 
	spring->value += ws->width ? ws->child->extra_width : ws->child->extra_height;
    }

  return spring->value;
}

static void
gtk_widget_spring_set_value (GtkSpring *spring,
			     gint       value)
{
  if (spring->value == GTK_SPRING_UNSET)
    {
      spring->min = GTK_SPRING_UNSET;
      spring->preferred = GTK_SPRING_UNSET;
      spring->max = GTK_SPRING_UNSET;
    }
  spring->value = value;
}

static void
gtk_widget_spring_finalize (GObject *object)
{
  GtkWidgetSpring *ws = GTK_WIDGET_SPRING (object);

  g_object_unref (ws->widget);
  ws->child = NULL;
}

static void
gtk_widget_spring_class_init (GtkWidgetSpringClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);
  GtkSpringClass *spring_class = GTK_SPRING_CLASS (class);

  object_class->finalize = gtk_widget_spring_finalize;

  spring_class->get_minimum_value = gtk_widget_spring_get_minimum;
  spring_class->get_preferred_value = gtk_widget_spring_get_preferred;
  spring_class->get_maximum_value = gtk_widget_spring_get_maximum;
  spring_class->get_value = gtk_widget_spring_get_value;
  spring_class->set_value = gtk_widget_spring_set_value;
}

static void
gtk_widget_spring_init (GtkWidgetSpring *spring)
{
}

static GType gtk_proxy_spring_get_type (void) G_GNUC_CONST;
#define GTK_TYPE_PROXY_SPRING (gtk_proxy_spring_get_type ())
#define GTK_PROXY_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_CAST ((object), GTK_TYPE_PROXY_SPRING, GtkProxySpring))
#define GTK_IS_PROXY_SPRING(object) \
  (G_TYPE_CHECK_INSTANCE_TYPE ((object), GTK_TYPE_PROXY_SPRING))

typedef struct _GtkProxySpring GtkProxySpring;
typedef GtkSpringClass GtkProxySpringClass;

struct _GtkProxySpring
{
  GtkSpring spring;
  
  GtkWidget *widget;

  GtkSpringLayoutChild *child;
  GtkConstraintType pos;
};

G_DEFINE_TYPE(GtkProxySpring, gtk_proxy_spring, GTK_TYPE_SPRING)

static gboolean
gtk_proxy_spring_ensure_child (GtkProxySpring *ps)
{
  if (ps->widget)
    {
      if (!GTK_IS_SPRING_LAYOUT (ps->widget->parent))
	ps->child = NULL;
      else if (!ps->child)
	ps->child = get_child (GTK_SPRING_LAYOUT (ps->widget->parent), 
			       ps->widget);
    }

  return ps->child != NULL;
}

static gint 
gtk_proxy_spring_get_minimum (GtkSpring *spring)
{
  GtkProxySpring *ps = GTK_PROXY_SPRING (spring);
  gint res = 0;
  
  if (gtk_proxy_spring_ensure_child (ps))
    {
      GtkSpring *s;

      s = gtk_spring_layout_child_get_constraint (ps->child, ps->pos);
      res = gtk_spring_get_minimum (s);
    }

  return res;
}

static gint 
gtk_proxy_spring_get_preferred (GtkSpring *spring)
{
  GtkProxySpring *ps = GTK_PROXY_SPRING (spring);
  gint res = 0;

  if (gtk_proxy_spring_ensure_child (ps))
    {
      GtkSpring *s;

      s = gtk_spring_layout_child_get_constraint (ps->child, ps->pos);
      res = gtk_spring_get_preferred (s);
    }

  return res;
}

static gint 
gtk_proxy_spring_get_maximum (GtkSpring *spring)
{
  GtkProxySpring *ps = GTK_PROXY_SPRING (spring);
  gint res = 0; /* FIXME */
  
  if (gtk_proxy_spring_ensure_child (ps))
    {
      GtkSpring *s;

      s = gtk_spring_layout_child_get_constraint (ps->child, ps->pos);
      res = gtk_spring_get_maximum (s);
    }

  return res;
}

static gint 
gtk_proxy_spring_get_value (GtkSpring *spring)
{
  GtkProxySpring *ps = GTK_PROXY_SPRING (spring);
  gint res = 0;

  if (gtk_proxy_spring_ensure_child (ps))
    {
      GtkSpring *s;

      s = gtk_spring_layout_child_get_constraint (ps->child, ps->pos);
      res = gtk_spring_get_value (s);
    }
  
  return res;
}

static void 
gtk_proxy_spring_set_value (GtkSpring *spring,
			    gint       value)
{
  GtkProxySpring *ps = GTK_PROXY_SPRING (spring);

  if (gtk_proxy_spring_ensure_child (ps)) 
    {
      GtkSpring *s;
      
      s = gtk_spring_layout_child_get_constraint (ps->child, ps->pos);
      gtk_spring_set_value (s, value);
    }
}

static gboolean 
gtk_proxy_spring_is_cyclic (GtkSpring *spring)
{
  GtkProxySpring *ps = GTK_PROXY_SPRING (spring);
  gboolean res = FALSE;

  if (gtk_proxy_spring_ensure_child (ps))
    {
      GtkSpring *s;
      
      s = gtk_spring_layout_child_get_constraint (ps->child, ps->pos);
      res = gtk_spring_test_cycle (s);
    }

  return res;
}

static void 
gtk_proxy_spring_reset_cyclic (GtkSpring *spring)
{
  GtkProxySpring *ps = GTK_PROXY_SPRING (spring);

  if (gtk_proxy_spring_ensure_child (ps))
    {
      GtkSpring *s;
      
      s = gtk_spring_layout_child_get_constraint (ps->child, ps->pos);
      gtk_spring_reset_cyclic (s);
    }
}

static void
gtk_proxy_spring_finalize (GObject *object)
{
  GtkProxySpring *ps = GTK_PROXY_SPRING (object);

  if (ps->widget)
    g_object_unref (ps->widget);
  ps->child = NULL;
}

static void
gtk_proxy_spring_class_init (GtkWidgetSpringClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);
  GtkSpringClass *spring_class = GTK_SPRING_CLASS (class);

  object_class->finalize = gtk_proxy_spring_finalize;

  spring_class->get_minimum_value = gtk_proxy_spring_get_minimum;
  spring_class->get_preferred_value = gtk_proxy_spring_get_preferred;
  spring_class->get_maximum_value = gtk_proxy_spring_get_maximum;
  spring_class->get_value = gtk_proxy_spring_get_value;
  spring_class->set_value = gtk_proxy_spring_set_value;
  spring_class->is_cyclic = gtk_proxy_spring_is_cyclic;
  spring_class->reset_cyclic = gtk_proxy_spring_reset_cyclic;
}

static void
gtk_proxy_spring_init (GtkProxySpring *spring)
{
}

GtkSpring *
gtk_spring_sink (GtkSpring *spring)
{
  if (spring->floating)
    {
      spring->floating = FALSE;
      g_object_unref (spring);
    }

  return spring;
}

gint       
gtk_spring_get_minimum (GtkSpring *spring)
{
  return GTK_SPRING_GET_CLASS (spring)->get_minimum_value (spring);
}

gint
gtk_spring_get_preferred (GtkSpring *spring)
{
  return GTK_SPRING_GET_CLASS (spring)->get_preferred_value (spring);
}

gint
gtk_spring_get_maximum (GtkSpring *spring)
{
  return GTK_SPRING_GET_CLASS (spring)->get_maximum_value (spring);
}

gint
gtk_spring_get_value (GtkSpring *spring)
{
  return GTK_SPRING_GET_CLASS (spring)->get_value (spring);
}

void
gtk_spring_set_value (GtkSpring *spring,
		      gint       value)
{
  if (!spring)
    return;

  GTK_SPRING_GET_CLASS (spring)->set_value (spring, value);
}

gboolean 
gtk_spring_is_cyclic (GtkSpring *spring)
{
  return GTK_SPRING_GET_CLASS (spring)->is_cyclic (spring);
}

void 
gtk_spring_reset_cyclic (GtkSpring *spring)
{
  if (!spring || spring->cyclic == GTK_SPRING_CYCLE_UNSET)
    return;

  spring->cyclic = GTK_SPRING_CYCLE_UNSET;
  
  if (GTK_SPRING_GET_CLASS (spring)->reset_cyclic)
    GTK_SPRING_GET_CLASS (spring)->reset_cyclic (spring);
}


GtkSpring *
gtk_spring_simple (gint min, 
		   gint preferred, 
		   gint max)
{
  GtkSpring *spring;

  spring = g_object_new (GTK_TYPE_SIMPLE_SPRING, NULL);

  spring->min = min;
  spring->preferred = preferred;
  spring->max = max;

  return spring;
}

GtkSpring *
gtk_spring_constant (gint c)
{
  return gtk_spring_simple (c, c, c);
}

#define ref_and_sink(spring) \
  gtk_spring_sink (GTK_SPRING (g_object_ref (spring)));

GtkSpring *
gtk_spring_scale (GtkSpring *spring1,
		  gdouble    factor)
{
  GtkSpring *spring;

  spring = g_object_new (GTK_TYPE_SCALE_SPRING, NULL);
  
  GTK_SCALE_SPRING (spring)->spring1 = ref_and_sink (spring1);
  GTK_SCALE_SPRING (spring)->factor = factor;

  return spring;
}

GtkSpring *
gtk_spring_minus (GtkSpring *spring1)
{
  return gtk_spring_scale (spring1, -1.0);
}

GtkSpring *
gtk_spring_sum (GtkSpring *spring1, 
		GtkSpring *spring2)
{
  GtkSpring *spring;

  spring = g_object_new (GTK_TYPE_COMBINE_SPRING, NULL);

  GTK_COMBINE_SPRING (spring)->spring1 = ref_and_sink (spring1);
  GTK_COMBINE_SPRING (spring)->spring2 = ref_and_sink (spring2);

  GTK_COMBINE_SPRING (spring)->op = GTK_COMBINE_SPRING_ADD;

  return spring;
}

GtkSpring *
gtk_spring_difference (GtkSpring *spring1, 
		       GtkSpring *spring2)
{
  GtkSpring *spring;

  spring = g_object_new (GTK_TYPE_COMBINE_SPRING, NULL);

  GTK_COMBINE_SPRING (spring)->spring1 = ref_and_sink (spring1);
  GTK_COMBINE_SPRING (spring)->spring2 = ref_and_sink (spring2);

  GTK_COMBINE_SPRING (spring)->op = GTK_COMBINE_SPRING_DIFFERENCE;

  return spring;
}

GtkSpring *
gtk_spring_max (GtkSpring *spring1, 
		GtkSpring *spring2)
{
  GtkSpring *spring;

  spring = g_object_new (GTK_TYPE_COMBINE_SPRING, NULL);

  GTK_COMBINE_SPRING (spring)->spring1 = ref_and_sink (spring1);
  GTK_COMBINE_SPRING (spring)->spring2 = ref_and_sink (spring2);

  GTK_COMBINE_SPRING (spring)->op = GTK_COMBINE_SPRING_MAX;

  return spring;
}

GtkSpring *
gtk_spring_width (GtkWidget *widget)
{
  GtkSpring *spring;

  spring = g_object_new (GTK_TYPE_WIDGET_SPRING, NULL);

  GTK_WIDGET_SPRING (spring)->widget = g_object_ref (widget);
  GTK_WIDGET_SPRING (spring)->width = TRUE;

  return spring;
}

GtkSpring *
gtk_spring_height (GtkWidget *widget)
{
  GtkSpring *spring;

  spring = g_object_new (GTK_TYPE_WIDGET_SPRING, NULL);

  GTK_WIDGET_SPRING (spring)->widget = g_object_ref (widget);
  GTK_WIDGET_SPRING (spring)->width = FALSE;

  return spring;
}

GtkSpring *
gtk_spring_proxy (GtkWidget         *widget, 
		  GtkConstraintType  pos)
{
  GtkSpring *spring;

  spring = g_object_new (GTK_TYPE_PROXY_SPRING, NULL);

  GTK_PROXY_SPRING (spring)->widget = g_object_ref (widget);
  GTK_PROXY_SPRING (spring)->pos = pos;

  return spring;
}

static void
name_constraints (GtkSpringLayoutChild *child, 
		  gchar                *pre, 
		  gchar                *post)
{
  g_print ("%s%s (%p)%s",
	   pre,
	   child && child->widget ? 
	   g_type_name (G_OBJECT_TYPE (G_OBJECT (child->widget)))
	   : "(NULL)",
	   child ? child->widget : NULL,
	   post);
}

static void
print_edge (GtkConstraintType pos)
{
  gchar *names[] = { "LEFT", "TOP", 
		     "RIGHT", "BOTTOM", 
		     "WIDTH", "HEIGHT", 
		     "UNKNOWN" }; 

  if (pos < GTK_CONSTRAINT_LEFT || pos > GTK_CONSTRAINT_HEIGHT)
    pos = GTK_CONSTRAINT_HEIGHT + 1;

  g_print ("%s", names[pos]);
}

static void
print_value (gint value)
{
  if (value == GTK_SPRING_UNSET)
    g_print ("UNSET");
  else
    g_print ("%d", value);
}

static void
dump_spring (GtkSpring *spring, 
	     gint       indent)
{
  if (!spring)
    {
      g_print ("%*s(NULL)\n", 2*indent, "");
      return;
    }

  g_print ("%*sspring %p ",
	   2*indent, "", spring);
  if (gtk_spring_is_cyclic (spring))
    g_print ("(cyclic) ");

  g_print ("[");
  print_value (spring->min);
  g_print(",");
  print_value (spring->preferred);
  g_print(",");
  print_value (spring->max);
  g_print("] ");
  print_value (spring->value);
  g_print(" ");

  if (GTK_IS_SIMPLE_SPRING (spring))
    {
      g_print ("\n");
    }
  else if (GTK_IS_COMBINE_SPRING (spring))
    {
      GtkCombineSpring *cs = GTK_COMBINE_SPRING (spring);
      
      switch (cs->op)
	{
	case GTK_COMBINE_SPRING_ADD:
	  g_print ("+\n");
	  break;
	case GTK_COMBINE_SPRING_MAX:
	  g_print ("M\n");
	  break;
	case GTK_COMBINE_SPRING_DIFFERENCE:
	  g_print ("-\n");
	  break;
	}	  
      dump_spring (cs->spring1, indent + 2);
      dump_spring (cs->spring2, indent + 2);
    }
  else if (GTK_IS_SCALE_SPRING (spring))
    {
      GtkScaleSpring *s = GTK_SCALE_SPRING (spring);
      
      g_print ("S %f\n", s->factor);
      dump_spring (s->spring1, indent + 2);
    }
  else if (GTK_IS_PROXY_SPRING (spring))
    {
      GtkProxySpring *ps = GTK_PROXY_SPRING (spring);
      GtkSpring *s;
      
      name_constraints (ps->child, "P ", " ");
      print_edge (ps->pos);
      g_print ("\n");
      s = gtk_spring_layout_child_get_constraint (ps->child, ps->pos);
      dump_spring (s, indent + 2);
    }
  else if (GTK_IS_WIDGET_SPRING (spring))
    {
      GtkWidgetSpring *ws = GTK_WIDGET_SPRING (spring);
      if (ws->width)
	name_constraints (ws->child, "W ", "\n");
      else
	name_constraints (ws->child, "H ", "\n");
    }
  else 
    {
      g_print ("unknown spring type %s\n", g_type_name (G_OBJECT_TYPE (G_OBJECT (spring))));
    }
}

static void
dump_constraints (GtkSpringLayoutChild *child)
{
  name_constraints (child, "LEFT for ", "\n");
  dump_spring (child->x, 2);

  name_constraints (child, "TOP for ", "\n");
  dump_spring (child->y, 2);

  name_constraints (child, "WIDTH for ", "\n");
  dump_spring (child->width, 2);

  name_constraints (child, "HEIGHT for ", "\n");
  dump_spring (child->height, 2);

  name_constraints (child, "RIGHT for ", "\n");
  dump_spring (child->right, 2);

  name_constraints (child, "BOTTOM for ", "\n");
  dump_spring (child->bottom, 2);

  name_constraints (child, "H for ", "\n");
  dump_spring (child->h, 2);

  name_constraints (child, "V for ", "\n");
  dump_spring (child->v, 2);
}

void
gtk_spring_layout_dump (GtkSpringLayout *layout)
{
  GList *children;
  GtkSpringLayoutChild *child;

  dump_constraints (layout->constraints);

  for (children = layout->children; children; children = children->next)
    {
      child = children->data;
      dump_constraints (child);
    }
}

/* GtkSpringLayoutChild */

#define ASSIGN(loc, obj)        \
  {                             \
    if (loc)		        \
      {                         \
        g_object_unref (loc);   \
        loc = NULL;             \
      }                         \
    if (obj)                    \
     {				\
       g_object_ref (obj);      \
       gtk_spring_sink (obj);   \
       loc = obj;               \
     }                          \
  }

GtkSpringLayoutChild *
gtk_spring_layout_child_new (void)
{
  GtkSpringLayoutChild *child;

  child = g_new0 (GtkSpringLayoutChild, 1);
}

GtkSpring *
gtk_spring_layout_child_get_x (GtkSpringLayoutChild *child)
{
  if (child->x)
    return child->x;

  if (!child->h && child->width && child->right)
    ASSIGN (child->h, gtk_spring_difference (child->right, child->width));

  return child->h;
}

GtkSpring *
gtk_spring_layout_child_get_y (GtkSpringLayoutChild *child)
{
  if (child->y)
    return child->y;

  if (!child->v && child->height && child->bottom)
    ASSIGN (child->v, gtk_spring_difference (child->bottom, child->height));

  return gtk_spring_sink (child->v);
}


GtkSpring *
gtk_spring_layout_child_get_width (GtkSpringLayoutChild *child)
{
  if (child->width)
    return child->width;
  
  if (!child->h && child->x && child->right)
    ASSIGN (child->h, gtk_spring_difference (child->right, child->x));

  return child->h;
}

GtkSpring *
gtk_spring_layout_child_get_height (GtkSpringLayoutChild *child)
{
  if (child->height)
    return child->height;
  else if (!child->v && child->y && child->bottom)
    ASSIGN (child->v, gtk_spring_difference (child->bottom, child->y));

  return child->v;
}

GtkSpring *
gtk_spring_layout_child_get_right (GtkSpringLayoutChild *child)
{
  if (child->right)
    return child->right;
  else if (!child->h && child->x && child->width)
    ASSIGN (child->h, gtk_spring_sum (child->x, child->width));
  
  return child->h;
}

GtkSpring *
gtk_spring_layout_child_get_bottom (GtkSpringLayoutChild *child)
{
  if (child->bottom)
    return child->bottom;
  else if (!child->v && child->y && child->height)
    ASSIGN (child->v, gtk_spring_sum (child->y, child->height));

  return child->v;
}

void
gtk_spring_layout_child_set_x (GtkSpringLayoutChild *child,
			       GtkSpring            *spring)
{
  ASSIGN (child->x, spring);
  ASSIGN (child->h, NULL);
  if (child->x && child->width && child->right)
    ASSIGN (child->width, NULL);
}

void
gtk_spring_layout_child_set_y (GtkSpringLayoutChild *child,
			       GtkSpring            *spring)
{
  ASSIGN (child->y, spring);
  ASSIGN (child->v, NULL);
  if (child->y && child->height && child->bottom)
    ASSIGN (child->height, NULL);
}

void
gtk_spring_layout_child_set_width (GtkSpringLayoutChild *child,
				    GtkSpring            *spring)
{
  ASSIGN (child->width, spring);
  ASSIGN (child->h, NULL);
  if (child->x && child->width && child->right)
    ASSIGN (child->right, NULL);
}

void
gtk_spring_layout_child_set_height (GtkSpringLayoutChild *child,
				    GtkSpring            *spring)
{
  ASSIGN (child->height, spring);
  ASSIGN (child->v, NULL);
  if (child->y && child->height && child->bottom)
    ASSIGN (child->bottom, NULL);
}

void
gtk_spring_layout_child_set_right (GtkSpringLayoutChild *child,
				   GtkSpring            *spring)
{
  ASSIGN (child->right, spring);
  ASSIGN (child->h, NULL);
  if (child->x && child->width && child->right)
    ASSIGN (child->x, NULL);
}

void
gtk_spring_layout_child_set_bottom (GtkSpringLayoutChild *child,
				    GtkSpring            *spring)
{
  ASSIGN (child->bottom, spring);
  ASSIGN (child->v, NULL);
  if (child->y && child->height && child->bottom)
    ASSIGN (child->y, NULL);
}

static GtkSpring *
gtk_spring_layout_child_get_constraint (GtkSpringLayoutChild *child,
					GtkConstraintType     pos)
{
  switch (pos)
    {
    case GTK_CONSTRAINT_LEFT:
      return gtk_spring_layout_child_get_x (child);
    case GTK_CONSTRAINT_TOP:
      return gtk_spring_layout_child_get_y (child);
    case GTK_CONSTRAINT_RIGHT:
      return gtk_spring_layout_child_get_right (child);
    case GTK_CONSTRAINT_BOTTOM:
      return gtk_spring_layout_child_get_bottom (child);
    case GTK_CONSTRAINT_WIDTH:
      return gtk_spring_layout_child_get_width (child);
    case GTK_CONSTRAINT_HEIGHT:
      return gtk_spring_layout_child_get_height (child);
    }
}

void 
gtk_spring_layout_child_set_constraint (GtkSpringLayoutChild *child,
					GtkConstraintType     pos,
					GtkSpring            *spring)
{
  switch (pos)
    {
    case GTK_CONSTRAINT_LEFT:
      gtk_spring_layout_child_set_x (child, spring);
      break;
    case GTK_CONSTRAINT_TOP:
      gtk_spring_layout_child_set_y (child, spring);
      break;
    case GTK_CONSTRAINT_RIGHT:
      gtk_spring_layout_child_set_right (child, spring);
      break;
    case GTK_CONSTRAINT_BOTTOM:
      gtk_spring_layout_child_set_bottom (child, spring);
      break;
    case GTK_CONSTRAINT_WIDTH:
      gtk_spring_layout_child_set_width (child, spring);
      break;
    case GTK_CONSTRAINT_HEIGHT:
      gtk_spring_layout_child_set_height (child, spring);
      break;
    }
}

static void
gtk_spring_layout_child_reset (GtkSpringLayoutChild *child)
{
  gtk_spring_set_value (child->x, GTK_SPRING_UNSET);
  gtk_spring_set_value (child->y, GTK_SPRING_UNSET);
  gtk_spring_set_value (child->width, GTK_SPRING_UNSET);
  gtk_spring_set_value (child->height, GTK_SPRING_UNSET);
  gtk_spring_set_value (child->right, GTK_SPRING_UNSET);
  gtk_spring_set_value (child->bottom, GTK_SPRING_UNSET);
  gtk_spring_set_value (child->h, GTK_SPRING_UNSET);
  gtk_spring_set_value (child->v, GTK_SPRING_UNSET);
}

static void
gtk_spring_layout_child_reset_cyclic (GtkSpringLayoutChild *child)
{
  gtk_spring_reset_cyclic (child->x);
  gtk_spring_reset_cyclic (child->y);
  gtk_spring_reset_cyclic (child->width);
  gtk_spring_reset_cyclic (child->height);
  gtk_spring_reset_cyclic (child->right);
  gtk_spring_reset_cyclic (child->bottom);
  gtk_spring_reset_cyclic (child->h);
  gtk_spring_reset_cyclic (child->v);
}

static void
gtk_spring_layout_reset (GtkSpringLayout *layout)
{
  GList *children;
  GtkSpringLayoutChild *child;

  gtk_spring_layout_child_reset_cyclic (layout->constraints);
  for (children = layout->children; children; children = children->next)
    {
      child = children->data;
      gtk_spring_layout_child_reset_cyclic (child);
    }

  gtk_spring_layout_child_set_x (layout->constraints, gtk_spring_constant (0));
  gtk_spring_layout_child_set_y (layout->constraints, gtk_spring_constant (0));
  gtk_spring_layout_child_set_width (layout->constraints, NULL);
  gtk_spring_layout_child_set_height (layout->constraints, NULL);
  if (!gtk_spring_layout_child_get_right (layout->constraints))
    gtk_spring_layout_child_set_right (layout->constraints, gtk_spring_simple (0, 0, 10000));
  if (!gtk_spring_layout_child_get_bottom (layout->constraints))
    gtk_spring_layout_child_set_bottom (layout->constraints, gtk_spring_simple (0, 0, 10000));
}

/* GtkSpringLayout */

enum {
  CHILD_PROP_0,
  CHILD_PROP_EXTRA_WIDTH,
  CHILD_PROP_MAX_EXTRA_WIDTH,
  CHILD_PROP_EXTRA_HEIGHT,
  CHILD_PROP_MAX_EXTRA_HEIGHT,
};

static void gtk_spring_layout_size_request  (GtkWidget        *widget,
					     GtkRequisition   *requisition);
static void gtk_spring_layout_size_allocate (GtkWidget        *widget,
					     GtkAllocation    *allocation);
static void gtk_spring_layout_add           (GtkContainer     *container,
					     GtkWidget        *widget);
static void gtk_spring_layout_remove        (GtkContainer     *container,
					     GtkWidget        *widget);
static void gtk_spring_layout_forall        (GtkContainer     *container,
					     gboolean 	       include_internals,
					     GtkCallback       callback,
					     gpointer          callback_data);
static GType gtk_spring_layout_child_type   (GtkContainer     *container);

static void gtk_spring_layout_set_child_property (GtkContainer *container,
						  GtkWidget    *child,
						  guint         property_id,
						  const GValue *value,
						  GParamSpec   *pspec);
static void gtk_spring_layout_get_child_property (GtkContainer *container,
						  GtkWidget    *child,
						  guint         property_id,
						  GValue       *value,
						  GParamSpec   *pspec);

G_DEFINE_TYPE(GtkSpringLayout, gtk_spring_layout, GTK_TYPE_CONTAINER)

static void
gtk_spring_layout_class_init (GtkSpringLayoutClass *class)
{
  GObjectClass *gobject_class;
  GtkObjectClass *gtkobject_class;
  GtkWidgetClass *widget_class;
  GtkContainerClass *container_class;

  gobject_class = G_OBJECT_CLASS (class);
  gtkobject_class = GTK_OBJECT_CLASS (class);
  widget_class = GTK_WIDGET_CLASS (class);
  container_class = GTK_CONTAINER_CLASS (class);

  widget_class->size_request = gtk_spring_layout_size_request;
  widget_class->size_allocate = gtk_spring_layout_size_allocate;

  container_class->add = gtk_spring_layout_add;
  container_class->remove = gtk_spring_layout_remove;
  container_class->forall = gtk_spring_layout_forall;
  container_class->child_type = gtk_spring_layout_child_type;

  container_class->set_child_property = gtk_spring_layout_set_child_property;
  container_class->get_child_property = gtk_spring_layout_get_child_property;

  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_EXTRA_WIDTH,
					      g_param_spec_int ("extra-width",
                                                                P_("Extra width"),
                                                                P_("Extra width to add to requisition"),
                                                                0,
                                                                G_MAXINT,
                                                                0,
                                                                G_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_MAX_EXTRA_WIDTH,
					      g_param_spec_int ("max-extra-width",
                                                                P_("Maximal extra width"),
                                                                P_("Maximal extra width to add to requisition"),
                                                                0,
                                                                G_MAXINT,
                                                                0,
                                                                G_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_EXTRA_HEIGHT,
					      g_param_spec_int ("extra-height",
                                                                P_("Extra height"),
                                                                P_("Extra height to add to requisition"),
                                                                0,
                                                                G_MAXINT,
                                                                0,
                                                                G_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
					      CHILD_PROP_MAX_EXTRA_WIDTH,
					      g_param_spec_int ("max-extra-height",
                                                                P_("Maximal extra height"),
                                                                P_("Maximal extra height to add to requisition"),
                                                                0,
                                                                G_MAXINT,
                                                                0,
                                                                G_PARAM_READWRITE));

}

static void
gtk_spring_layout_init (GtkSpringLayout *layout)
{
  GtkSpringLayoutChild *child;

  GTK_WIDGET_SET_FLAGS (layout, GTK_NO_WINDOW);
   
  child = g_new0 (GtkSpringLayoutChild, 1);
  child->widget = NULL;

  child->x = gtk_spring_constant (0);
  child->y = gtk_spring_constant (0);

#if 0
  child->width = gtk_spring_simple (1, 1, 10000);
  child->height = gtk_spring_simple (1, 1, 10000);
#endif

  layout->children = NULL;

  layout->constraints = child;

  layout->cycle = gtk_spring_constant (GTK_SPRING_UNSET);
}

static GType
gtk_spring_layout_child_type (GtkContainer *container)
{
  return GTK_TYPE_WIDGET;
}


GtkWidget*
gtk_spring_layout_new (void)
{
  return g_object_new (GTK_TYPE_SPRING_LAYOUT, NULL);
}

static GtkSpringLayoutChild*
get_child (GtkSpringLayout  *layout,
           GtkWidget        *widget)
{
  GList *children;
  
  if (widget == (GtkWidget *)layout)
    return layout->constraints;

  children = layout->children;
  while (children)
    {
      GtkSpringLayoutChild *child;
      
      child = children->data;
      children = children->next;
      
      if (child->widget == widget)
        return child;
    }
  
  return NULL;
}

static void
gtk_spring_layout_set_child_property (GtkContainer    *container,
				      GtkWidget       *child,
				      guint            property_id,
				      const GValue    *value,
				      GParamSpec      *pspec)
{
  GtkSpringLayoutChild *info;

  info = get_child (GTK_SPRING_LAYOUT (container), child);

  g_return_if_fail (info != NULL);

  switch (property_id)
    {
    case CHILD_PROP_EXTRA_WIDTH:
      info->extra_width = g_value_get_int (value);
      break;
    case CHILD_PROP_MAX_EXTRA_WIDTH:
      info->max_extra_width = g_value_get_int (value);
      break;
    case CHILD_PROP_EXTRA_HEIGHT:
      info->extra_height = g_value_get_int (value);
      break;
    case CHILD_PROP_MAX_EXTRA_HEIGHT:
      info->max_extra_height = g_value_get_int (value);
      break;
    default:
      GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
      break;
    }
}

static void
gtk_spring_layout_get_child_property (GtkContainer *container,
				      GtkWidget    *child,
				      guint         property_id,
				      GValue       *value,
				      GParamSpec   *pspec)
{
  GtkSpringLayoutChild *info;

  info = get_child (GTK_SPRING_LAYOUT (container), child);

  g_return_if_fail (info != NULL);
  
  switch (property_id)
    {
    case CHILD_PROP_EXTRA_WIDTH:
      g_value_set_int (value, info->extra_width);
      break;
    case CHILD_PROP_MAX_EXTRA_WIDTH:
      g_value_set_int (value, info->max_extra_width);
      break;
    case CHILD_PROP_EXTRA_HEIGHT:
      g_value_set_int (value, info->extra_height);
      break;
    case CHILD_PROP_MAX_EXTRA_HEIGHT:
      g_value_set_int (value, info->max_extra_height);
      break;
    default:
      GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
      break;
    }
}

GtkSpring *acyclic_spring (GtkSpringLayout *layout,
			   GtkSpring       *spring)
{
  if (gtk_spring_test_cycle (spring))
    return layout->cycle;
  else
    return spring;
}

static void
gtk_spring_layout_size_request (GtkWidget      *widget,
				GtkRequisition *requisition)
{
  GtkSpringLayout *layout;  
  GtkSpringLayoutChild *child;
  GList *children;
  GtkRequisition child_requisition;
  GtkSpring *x, *y, *width, *height;
  gint right, bottom;

  layout = GTK_SPRING_LAYOUT (widget);

  gtk_spring_layout_reset (layout);
  for (children = layout->children; children; children = children->next)
    {
      child = children->data;
      gtk_widget_size_request (child->widget, &child_requisition);
    }

  width = acyclic_spring (layout, gtk_spring_layout_child_get_width (layout->constraints));
  height = acyclic_spring (layout, gtk_spring_layout_child_get_height (layout->constraints));
  
  requisition->width = gtk_spring_get_preferred (width);
  requisition->height = gtk_spring_get_preferred (height);
}

static void
gtk_spring_layout_size_allocate (GtkWidget     *widget,
				 GtkAllocation *allocation)
{
  GtkSpringLayout *layout;
  GtkSpringLayoutChild *child;
  GtkAllocation child_allocation;
  GtkRequisition child_requisition;
  GList *children;
  GtkSpring *x, *y, *width, *height;

  layout = GTK_SPRING_LAYOUT (widget);

  widget->allocation = *allocation;

  gtk_spring_layout_reset (layout);
  gtk_spring_layout_child_reset (layout->constraints);
  for (children = layout->children; children; children = children->next)
    {
      child = children->data;
      gtk_spring_layout_child_reset (child);
    }

  child = layout->constraints;
  x = acyclic_spring (layout, gtk_spring_layout_child_get_x (child));
  y = acyclic_spring (layout, gtk_spring_layout_child_get_y (child));
  width = acyclic_spring (layout, gtk_spring_layout_child_get_width (child));
  height = acyclic_spring (layout, gtk_spring_layout_child_get_height (child));
  gtk_spring_set_value (x, 0);
  gtk_spring_set_value (y, 0);
  gtk_spring_set_value (width, allocation->width);
  gtk_spring_set_value (height, allocation->height);

  for (children = layout->children; children; children = children->next)
    {
      child = children->data;
      
      if (GTK_WIDGET_VISIBLE (child->widget))
	{
	  x = acyclic_spring (layout, gtk_spring_layout_child_get_x (child));
	  y = acyclic_spring (layout, gtk_spring_layout_child_get_y (child));
	  width = acyclic_spring (layout, gtk_spring_layout_child_get_width (child));
	  height = acyclic_spring (layout, gtk_spring_layout_child_get_height (child));
	  
	  child_allocation.x = widget->allocation.x;
	  child_allocation.y = widget->allocation.y;
	  child_allocation.width = widget->requisition.width;
	  child_allocation.height = widget->requisition.height;

	  child_allocation.x += gtk_spring_get_value (x);
	  child_allocation.y += gtk_spring_get_value (y);
	  child_allocation.width = gtk_spring_get_value (width);
	  child_allocation.height = gtk_spring_get_value (height);
	  
	  gtk_widget_size_allocate (child->widget, &child_allocation);
	}
    }
}

static void
gtk_spring_layout_add (GtkContainer *container,
		       GtkWidget    *widget)
{
  GtkSpringLayout *layout = GTK_SPRING_LAYOUT (container);
  GtkSpringLayoutChild *child;

  child = g_new0 (GtkSpringLayoutChild, 1);
  child->widget = widget;

  gtk_widget_set_parent (widget, GTK_WIDGET (layout));
  layout->children = g_list_append (layout->children, child);

  child->x = gtk_spring_constant (0);
  child->y = gtk_spring_constant (0);
  child->width = gtk_spring_width (widget);
  child->height = gtk_spring_height (widget);  
}

static void
gtk_spring_layout_remove (GtkContainer *container,
			  GtkWidget    *widget)
{
  GtkSpringLayout *layout;
  GtkSpringLayoutChild *child, *child2;
  GList *children;

  /* FIXME clean up here */
  layout = GTK_SPRING_LAYOUT (container);

  children = layout->children;
  while (children)
    {
      child = children->data;

      if (child->widget == widget)
	{
	  gboolean was_visible = GTK_WIDGET_VISIBLE (widget);
	  
	  gtk_widget_unparent (widget);

	  layout->children = g_list_remove_link (layout->children, children);
	  g_list_free (children);

	  ASSIGN (child->x, NULL);
	  ASSIGN (child->y, NULL);
	  ASSIGN (child->width, NULL);
	  ASSIGN (child->height, NULL);
	  ASSIGN (child->right, NULL);
	  ASSIGN (child->bottom, NULL);
	  ASSIGN (child->h, NULL);
	  ASSIGN (child->v, NULL);
	  
	  g_free (child);
	  
	  if (was_visible && GTK_WIDGET_VISIBLE (container))
	    gtk_widget_queue_resize (GTK_WIDGET (container));

	  break;
	}
      
      children = children->next;
    }
}

static void
gtk_spring_layout_forall (GtkContainer *container,
			  gboolean	include_internals,
			  GtkCallback   callback,
			  gpointer      callback_data)
{
  GtkSpringLayout *layout;
  GtkSpringLayoutChild *child;
  GList *children;

  g_return_if_fail (callback != NULL);

  layout = GTK_SPRING_LAYOUT (container);

  children = layout->children;
  while (children)
    {
      child = children->data;
      children = children->next;

      (* callback) (child->widget, callback_data);
    }
}


void
gtk_spring_layout_add_spring (GtkSpringLayout   *layout,
			      GtkWidget         *widget,
			      GtkConstraintType  pos,
			      gint               min,
			      gint               preferred,
			      gint               max,
			      GtkWidget         *target,
			      GtkConstraintType  target_pos)
{
  GtkSpringLayoutChild *child;
  GtkSpringLayoutChild *target_child;
  GtkSpring *spring, *target_spring, *sum;

  child = get_child (layout, widget);
  target_child = get_child (layout, target);
  
  g_return_if_fail (child && target_child);

  spring = gtk_spring_simple (min, preferred, max);
  target_spring = gtk_spring_proxy (target, target_pos);
  sum = gtk_spring_sum (spring, target_spring);

  gtk_spring_layout_child_set_constraint (child, pos, sum);
}

void
gtk_spring_layout_add_strut (GtkSpringLayout   *layout,
			     GtkWidget         *widget,
			     GtkConstraintType  pos,
			     gint               size,
			     GtkWidget         *target,
			     GtkConstraintType  target_pos)
{
  gtk_spring_layout_add_spring (layout, widget, pos, 
				size, size, size,
				target, target_pos);
}


void
gtk_spring_layout_set_constraint (GtkSpringLayout   *layout,
				  GtkWidget         *widget,
				  GtkConstraintType  pos,
				  GtkSpring         *spring)
{
  GtkSpringLayoutChild *child;

  child = get_child (layout, widget);

  g_return_if_fail (child);
  
  gtk_spring_layout_child_set_constraint (child, pos, spring);
}

GtkSpring *
gtk_spring_layout_get_constraint (GtkSpringLayout   *layout,
				  GtkWidget         *widget,
				  GtkConstraintType  pos)
{
  GtkSpringLayoutChild *child;

  child = get_child (layout, widget);

  g_return_if_fail (child);
  
  return gtk_spring_layout_child_get_constraint (child, pos);
}
